# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from tempfile import mkstemp
from threading import Event, Lock
from time import time
from typing import List

from neon_utils.file_utils import encode_file_to_base64_string
from ovos_utils.log import LOG
from ovos_bus_client.client import MessageBusClient
from ovos_bus_client.message import Message
from ovos_plugin_manager.tts import TTS


class UtteranceTests:
    def __init__(self, prompts: List[str], lang: str = "en-us",
                 bus_config: dict = None, user_config: dict = None,
                 audio: bool = False, tts: TTS = None):
        if not user_config:
            from neon_utils.configuration_utils import get_neon_user_config
            user_config = get_neon_user_config().content
        user_config['user']['username'] = "minerva"
        self._user_config = user_config
        bus_config = bus_config or dict()
        self.core_bus = MessageBusClient(**bus_config)
        self.core_bus.run_in_thread()
        self.lang = lang
        self.test_audio = audio
        self._tts = tts
        # TODO: Handle prompt metadata for longer timeouts
        self._prompts = prompts
        self._stt_timeout = 60     # Time to transcribe + audio parsers
        self._intent_timeout = 60  # Time to match AND handle intent
        self._speak_timeout = 60   # Time after intent handling for TTS playback

        self._results = list()
        self._audio_output_done = Event()
        self._prompt_handled = Event()
        self._prompt_lock = Lock()

        self._last_message = None
        self._audio_output_done.set()
        self.register_bus_events()

    def run_test(self) -> dict:
        """
        Run tests and return dict timing results
        """
        self._results.clear()
        for prompt in self._prompts:
            self.handle_prompt(prompt)
        aggregated_results = {"save_transcript": [],
                              "text_parsers": [],
                              "get_tts": [],
                              "intent_handler": [],
                              "total": []}
        if self.test_audio:
            aggregated_results['get_stt'] = []
        for result in self._results:
            try:
                aggregated_results['save_transcript'].append(result['save_transcript'])
                aggregated_results['text_parsers'].append(result['text_parsers'])
                aggregated_results['get_tts'].append(result['get_tts'])
                aggregated_results['intent_handler'].append(result['speech_start'] - result['handle_utterance'])
                aggregated_results['total'].append(result['finished'] - result['transcribed'])
                if self.test_audio:
                    aggregated_results['get_stt'].append(result['get_stt'])
            except KeyError:
                LOG.error(result)
        formatted_results = dict()
        for key, values in aggregated_results.items():
            formatted_results[key] = {"average": round(sum(values) /
                                                       len(values), 6),
                                      "minimum": round(min(values), 6),
                                      "maximum": round(max(values), 6)}
        return formatted_results

    def register_bus_events(self):
        """
        Register listeners to track audio and skill module states
        """
        self.core_bus.on("recognizer_loop:audio_output_start",
                         self._audio_started)
        self.core_bus.on("recognizer_loop:audio_output_end",
                         self._audio_stopped)
        self.core_bus.on("mycroft.mic.listen", self._mic_listen)
        self.core_bus.on("mycroft.skill.handler.complete",
                         self._handler_complete)

    def _audio_started(self, _):
        """
        Handle audio output started
        """
        self._audio_output_done.clear()

    def _audio_stopped(self, message):
        """
        Handle audio output finished
        @param message: Message associated with completed audio playback
        """
        LOG.debug("audio finished")
        self._last_message = message
        self._audio_output_done.set()

    def _mic_listen(self, message):
        """
        Handle start listening (for prompts that trigger `get_response`)
        @param message: Message associated with completed skill handler
        """
        LOG.debug("`get_response` call")
        # self._last_message = message
        self._prompt_handled.set()

    def _handler_complete(self, _):
        """
        Handle skill execution complete (audio output may not be complete)
        """
        LOG.debug("Skill Handler Complete")
        self._prompt_handled.set()

    def send_prompt(self, prompt: str):
        """
        Send a prompt to core for intent handling
        """
        context = {"neon_should_respond": True,
                   "source": ["minerva"],
                   "destination": ["skills"],
                   "timing": {"transcribed": time()},
                   "username": "minerva",
                   "user_profiles": [self._user_config]}
        if self.test_audio:
            if self._tts:
                _, file_path = mkstemp()
                audio, _ = self._tts.get_tts(prompt, file_path, lang=self.lang)
            else:
                resp = self.core_bus.wait_for_response(
                    Message("neon.get_tts", {'text': prompt,
                                             'speaker': {'language': self.lang,
                                                         'gender': 'female'}}),
                    timeout=self._stt_timeout)
                file_path = resp.data[self.lang]['female']
            resp = self.core_bus.wait_for_response(
                Message("neon.audio_input",
                        {"audio_data": encode_file_to_base64_string(file_path),
                         "lang": self.lang}, context),
                timeout=self._stt_timeout)
            LOG.info(resp.data)
            if prompt.lower() not in (t.lower() for t
                                      in resp.data['transcripts']):
                LOG.warning(f"Invalid transcription for '{prompt}': "
                            f"{resp.data['transcripts']}")
        else:
            self.core_bus.emit(Message("recognizer_loop:utterance",
                                       {"utterances": [prompt],
                                        "lang": self.lang}, context))

    def handle_prompt(self, prompt: str):
        """
        Send a prompt (text or audio) and collect timing results.
        @param prompt: string prompt to send for intent (and optionally STT)
            processing
        """
        with self._prompt_lock:
            # Ensure event state matches expectation
            if not self._audio_output_done.is_set():
                LOG.warning("Audio output not finished when expected!")
                self._audio_output_done.wait(self._speak_timeout)
            self._audio_output_done.clear()
            self._prompt_handled.clear()
            self._last_message = None

            # Send prompt
            self.send_prompt(prompt)
            try:
                assert self._prompt_handled.wait(self._intent_timeout)
                assert self._audio_output_done.wait(self._speak_timeout)
                assert self._last_message is not None
                if "speech_start" not in self._last_message.context["timing"]:
                    LOG.warning(f"Missing speech_start timestamp for {prompt}")
                    self._last_message.context["timing"]["speech_start"] = \
                        self._last_message.context["timing"]["handle_utterance"]
                self._results.append({**self._last_message.context["timing"],
                                      **{'finished': time()}})
            except AssertionError as e:
                LOG.error(f"{prompt}: {e}")
        LOG.debug(f"Handled {prompt}")



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
import shutil

from os import environ, getenv
from os.path import dirname, join
from unittest.mock import Mock
from ovos_utils.messagebus import FakeBus

from neon_minerva.skill import get_skill_object


class SkillTestCase(unittest.TestCase):
    # Define test directories
    test_fs = join(dirname(__file__), "skill_fs")
    data_dir = join(test_fs, "data")
    conf_dir = join(test_fs, "config")
    environ["XDG_DATA_HOME"] = data_dir
    environ["XDG_CONFIG_HOME"] = conf_dir

    # Define static parameters
    bus = FakeBus()
    bus.run_forever()
    test_skill_id = 'test_skill.test'

    skill = None

    @classmethod
    def setUpClass(cls) -> None:
        # Get test skill
        skill_entrypoint = getenv("TEST_SKILL_ENTRYPOINT")
        if not skill_entrypoint:
            from ovos_plugin_manager.skills import find_skill_plugins
            skill_entrypoints = list(find_skill_plugins().keys())
            assert len(skill_entrypoints) == 1
            skill_entrypoint = skill_entrypoints[0]

        cls.skill = get_skill_object(skill_entrypoint=skill_entrypoint,
                                     skill_id=cls.test_skill_id, bus=cls.bus)
        # Override speak and speak_dialog to test passed arguments
        cls.skill.speak = Mock()
        cls.skill.speak_dialog = Mock()

    def setUp(self):
        self.skill.speak.reset_mock()
        self.skill.speak_dialog.reset_mock()

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree(cls.test_fs)



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2021 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import yaml

from os.path import expanduser, isfile, isdir
from typing import Optional
from ovos_utils.messagebus import FakeBus
from ovos_workshop.skills.base import BaseSkill
from ovos_utils.log import LOG


def get_skill_object(skill_entrypoint: str, bus: FakeBus,
                     skill_id: str, config_patch: Optional[dict] = None) -> BaseSkill:
    """
    Get an initialized skill object by entrypoint with the requested skill_id.
    @param skill_entrypoint: Skill plugin entrypoint or directory path
    @param bus: FakeBus instance to bind to skill for testing
    @param skill_id: skill_id to initialize skill with
    @returns: Initialized skill object
    """
    if config_patch:
        from ovos_config.config import update_mycroft_config
        update_mycroft_config(config_patch)
    if isdir(skill_entrypoint):
        LOG.info(f"Loading local skill: {skill_entrypoint}")
        from ovos_workshop.skill_launcher import SkillLoader
        loader = SkillLoader(bus, skill_entrypoint, skill_id)
        if loader.load():
            return loader.instance
    from ovos_plugin_manager.skills import find_skill_plugins
    plugins = find_skill_plugins()
    if skill_entrypoint not in plugins:
        raise ValueError(f"Requested skill not found: {skill_entrypoint}")
    plugin = plugins[skill_entrypoint]
    skill = plugin(bus=bus, skill_id=skill_id)
    return skill


def load_resource_tests(test_file: str) -> dict:
    """
    Load resource tests from a file
    @param test_file: Test file to load
    @returns: Loaded test spec
    """
    test_file = expanduser(test_file)
    if not isfile(test_file):
        raise FileNotFoundError(test_file)
    with open(test_file) as f:
        resources = yaml.safe_load(f)
    return resources


def load_intent_tests(test_file: str) -> dict:
    """
    Load intent tests from a file
    @param test_file: Test file to load
    @returns: Loaded test spec
    """
    test_file = expanduser(test_file)
    if not isfile(test_file):
        raise FileNotFoundError(test_file)
    with open(test_file) as f:
        intents = yaml.safe_load(f)
    return intents
